/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check.error;

import java.util.List;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ErrorReport;
import de.hft.stuttgart.citydoctor2.check.ErrorType;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GmlElement;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

/**
 * A vertex which is contained in multiple polygons which are not connected via
 * edges.
 * 
 * @author Matthias Betz
 *
 */
public class NonManifoldVertexError implements CheckError {

	private static final long serialVersionUID = -3915669943428175777L;
	
	private List<List<Polygon>> components;
	private Geometry geom;
	private Vertex v;

	public NonManifoldVertexError(Geometry geom, List<List<Polygon>> components, Vertex v) {
		this.geom = geom;
		this.components = components;
		this.v = v;
	}

	public List<List<Polygon>> getComponents() {
		return components;
	}

	public Geometry getGeometry() {
		return geom;
	}

	public Vertex getVertex() {
		return v;
	}

	@Override
	public void report(ErrorReport report) {
		report.add(geom);
		report.add(v);
		for (int i = 0; i < components.size(); i++) {
			List<Polygon> polygons = components.get(i);
			report.add("component nr: ", i);
			for (Polygon p : polygons) {
				report.add(p);
			}
		}

	}

	@Override
	public void accept(ErrorVisitor errorVisitor) {
		errorVisitor.visit(this);
	}

	@Override
	public boolean accept(HealingMethod method, ModificationListener l) {
		return method.visit(this, l);
	}

	@Override
	public String toString() {
		return "NonManifoldVertexError [components=" + components + ", geom=" + geom + ", v=" + v + "]";
	}

	@Override
	public ErrorType getType() {
		return ErrorType.ERROR;
	}

	@Override
	public ErrorId getErrorId() {
		return ErrorId.GE_S_NON_MANIFOLD_VERTEX;
	}

	@Override
	public GmlElement getFeature() {
		return getGeometry();
	}
}
