/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check.error;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ErrorReport;
import de.hft.stuttgart.citydoctor2.check.ErrorType;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.datastructure.GmlElement;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.Plane;
import de.hft.stuttgart.citydoctor2.utils.Localization;

/**
 * Error object for planarity errors containing the polygon, the regression
 * plane and the distance to the plane.
 * 
 * @author Matthias Betz
 *
 */
public class NonPlanarPolygonDistancePlaneError implements CheckError {

	private static final long serialVersionUID = -3504364055236383519L;
	
	private Polygon p;
	private double distance;
	private Vertex v;
	private Plane plane;

	public NonPlanarPolygonDistancePlaneError(Polygon p, double distance, Vertex v, Plane plane) {
		this.p = p;
		this.distance = distance;
		this.v = v;
		this.plane = plane;
	}

	public Plane getPlane() {
		return plane;
	}

	public Polygon getPolygon() {
		return p;
	}

	public double getDistance() {
		return distance;
	}

	public Vertex getVertex() {
		return v;
	}

	@Override
	public void accept(ErrorVisitor errorVisitor) {
		errorVisitor.visit(this);
	}

	@Override
	public boolean accept(HealingMethod method, ModificationListener l) {
		return method.visit(this, l);
	}

	@Override
	public void report(ErrorReport report) {
		report.add(p);
		report.add(v);
		report.add(Localization.getText("DistanceError.distanceFromPlane"), distance);
	}

	@Override
	public String toString() {
		return "NonPlanarPolygonDistancePlaneError [p=" + p + ", distance=" + distance + ", v=" + v + ", plane=" + plane + "]";
	}

	@Override
	public ErrorType getType() {
		return ErrorType.ERROR;
	}

	@Override
	public ErrorId getErrorId() {
		return ErrorId.GE_P_NON_PLANAR_POLYGON_DISTANCE_PLANE;
	}

	@Override
	public GmlElement getFeature() {
		return getPolygon();
	}
}
