/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check.error;

import java.util.List;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ErrorReport;
import de.hft.stuttgart.citydoctor2.check.ErrorType;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.datastructure.GmlElement;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;

/**
 * If the holes in a polygon split the interior area into two distinct areas
 * this error is created.
 * 
 * @author Matthias Betz
 *
 */
public class PolygonInteriorDisconnectedError implements CheckError {

	private static final long serialVersionUID = 511849016699842395L;
	
	private Polygon p;
	private List<LinearRing> connectedRings;

	public PolygonInteriorDisconnectedError(Polygon p, List<LinearRing> connectedRings) {
		this.p = p;
		this.connectedRings = connectedRings;
	}

	public Polygon getPolygon() {
		return p;
	}

	public List<LinearRing> getConnectedRings() {
		return connectedRings;
	}

	@Override
	public void report(ErrorReport report) {
		report.add(p);
		report.add("connected rings", "");
		for (LinearRing lr : connectedRings) {
			report.add(lr);
		}
	}

	@Override
	public void accept(ErrorVisitor errorVisitor) {
		errorVisitor.visit(this);
	}

	@Override
	public boolean accept(HealingMethod method, ModificationListener l) {
		return method.visit(this, l);
	}

	@Override
	public String toString() {
		return "InteriorDisconnectedError [p=" + p + ", connectedRings=" + connectedRings + "]";
	}

	@Override
	public ErrorType getType() {
		return ErrorType.ERROR;
	}

	@Override
	public ErrorId getErrorId() {
		return ErrorId.GE_P_INTERIOR_DISCONNECTED;
	}

	@Override
	public GmlElement getFeature() {
		return getPolygon();
	}
}
