/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check.error;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ErrorReport;
import de.hft.stuttgart.citydoctor2.check.ErrorType;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.GmlElement;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.math.Vector3d;

/**
 * When two edges of the same linear ring are intersecting with each other this
 * error object is created.
 * 
 * @author Matthias Betz
 *
 */
public class RingEdgeIntersectionError implements CheckError {

	private static final long serialVersionUID = 4923311259469198172L;

	private Edge e1;
	private Edge e2;
	private Vector3d intersection;
	private LinearRing lr;

	public RingEdgeIntersectionError(LinearRing lr, Edge e1, Edge e2, Vector3d intersection) {
		if (e1 == null || e2 == null || lr == null || intersection == null) {
			throw new IllegalArgumentException("Parameters may not be null");
		}
		this.e1 = e1;
		this.e2 = e2;
		this.intersection = intersection;
		this.lr = lr;
	}

	public LinearRing getRing() {
		return lr;
	}

	public Edge getEdge1() {
		return e1;
	}

	public Edge getEdge2() {
		return e2;
	}

	public Vector3d getIntersection() {
		return intersection;
	}

	@Override
	public void report(ErrorReport report) {
		report.add("type", "edge intersection");
		report.add(lr);
		report.add(e1);
		report.add(e2);
		report.add("intersection point", intersection);
	}

	@Override
	public void accept(ErrorVisitor errorVisitor) {
		errorVisitor.visit(this);
	}

	@Override
	public boolean accept(HealingMethod method, ModificationListener l) {
		return method.visit(this, l);
	}

	@Override
	public String toString() {
		return "EdgeIntersectionError [e1=" + e1 + ", e2=" + e2 + ", intersection=" + intersection + ", lr=" + lr + "]";
	}

	@Override
	public ErrorType getType() {
		return ErrorType.ERROR;
	}

	@Override
	public ErrorId getErrorId() {
		return ErrorId.GE_R_SELF_INTERSECTION;
	}

	@Override
	public GmlElement getFeature() {
		return getRing();
	}
}
