/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check.error;

import java.util.List;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ErrorReport;
import de.hft.stuttgart.citydoctor2.check.ErrorType;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GmlElement;

/**
 * If a geometry has holes in it, it results in edges being traversed only once.
 * This error list all those edges.
 * 
 * @author Matthias Betz
 *
 */
public class SolidNotClosedError implements CheckError {

	private static final long serialVersionUID = 8256385219885788891L;
	
	private List<Edge> errorEdges;
	private Geometry g;

	public SolidNotClosedError(Geometry g, List<Edge> errorEdges) {
		this.errorEdges = errorEdges;
		this.g = g;
	}

	public Geometry getGeometry() {
		return g;
	}

	public List<Edge> getErrorEdges() {
		return errorEdges;
	}

	@Override
	public void report(ErrorReport report) {
		report.add(g);
		report.addEdges("open edges", errorEdges);
	}

	@Override
	public void accept(ErrorVisitor errorVisitor) {
		errorVisitor.visit(this);
	}

	@Override
	public boolean accept(HealingMethod method, ModificationListener l) {
		return method.visit(this, l);
	}

	@Override
	public String toString() {
		return "SolidNotClosedError [errorEdges=" + errorEdges + ", g=" + g + "]";
	}

	@Override
	public ErrorType getType() {
		return ErrorType.ERROR;
	}

	@Override
	public ErrorId getErrorId() {
		return ErrorId.GE_S_NOT_CLOSED;
	}

	@Override
	public GmlElement getFeature() {
		return getGeometry();
	}
}
