/*-
 *  Copyright 2022 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.datastructure;

import java.util.ArrayList;
import java.util.List;

import org.citygml4j.core.model.relief.ReliefFeature;
import org.citygml4j.core.util.geometry.GeometryFactory;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;

public class ReliefObject extends CityObject {
	
	private static final long serialVersionUID = -9162169874426519903L;

	private ReliefFeature feature;
	
	private List<TinObject> components = new ArrayList<>();
	
	public ReliefObject(ReliefFeature feature) {
		this.feature = feature;
	}
	
	@Override
	public void accept(Check c) {
		super.accept(c);
		if (c.canExecute(this)) {
			c.check(this);
		}
		for (TinObject tin : components) {
			tin.accept(c);
		}
	}

	@Override
	public void collectInstances(CopyHandler handler) {
		handler.addInstance(components);
	}
	
	@Override
	public void fillValues(Copyable original, CopyHandler handler) {
		super.fillValues(original, handler);
		ReliefObject originalRelief = (ReliefObject) original;
		for (TinObject tin : originalRelief.components) {
			components.add(handler.getCopyInstance(tin));
		}
	}
	
	@Override
	public Copyable createCopyInstance() {
		return new ReliefObject(feature);
	}

	@Override
	public void prepareForChecking() {
		for (TinObject component : components) {
			component.prepareForChecking();
		}
	}

	@Override
	public void clearMetaInformation() {
		for (TinObject component : components) {
			component.clearMetaInformation();
		}		
	}

	@Override
	public void clearAllContainedCheckResults() {
		for (TinObject component : components) {
			component.clearAllContainedCheckResults();
		}			
	}
	
	@Override
	public void collectContainedErrors(List<CheckError> errors) {
		super.collectContainedErrors(errors);
		for (TinObject tin : components) {
			tin.collectContainedErrors(errors);
		}
	}
	
	@Override
	public boolean containsAnyError() {
		boolean containsError = super.containsAnyError();
		if (containsError) {
			return true;
		}
		for (TinObject tin : components) {
			if (tin.containsAnyError()) {
				return true;
			}
		}
		return false;
	}
	
	@Override
	public boolean containsError(CheckId checkIdentifier) {
		boolean hasError = super.containsError(checkIdentifier);
		if (hasError) {
			return true;
		}
		for (TinObject tin : components) {
			if (tin.containsError(checkIdentifier)) {
				return true;
			}
		}
		return false;
	}

	@Override
	public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
		// no geometries
	}

	@Override
	public ReliefFeature getGmlObject() {
		return feature;
	}

	@Override
	public void unsetGmlGeometries() {
		// no geometries
	}

	@Override
	public FeatureType getFeatureType() {
		return FeatureType.LAND;
	}
	
	public List<TinObject> getComponents() {
		return components;
	}

}
