/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.RequirementType;
import de.hft.stuttgart.citydoctor2.check.Checkable;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.error.ConsecutivePointSameError;
import de.hft.stuttgart.citydoctor2.check.error.RingDuplicatePointError;
import de.hft.stuttgart.citydoctor2.checks.util.CollectionUtils;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;

/**
 * CP_DUPPOINT checks if there is any double point in a linear ring. Each point
 * must occur only once in a linear ring. <br/>
 * <br/>
 * <b>Dependency:</b> <br/>
 * <dd>
 * <dd>CP_CLOSE</dd></dd>
 * 
 * @author Matthias Betz
 * 
 */
public class DuplicatePointsCheck extends Check {

	private static final String EPSILON_NAME = "minVertexDistance";

	private static final List<CheckId> dependencies;

	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		deps.add(CheckId.C_GE_R_TOO_FEW_POINTS);
		deps.add(CheckId.C_GE_R_NOT_CLOSED);
		dependencies = Collections.unmodifiableList(deps);

		ArrayList<Class<? extends Checkable>> classes = new ArrayList<>();
		classes.add(LinearRing.class);
	}

	private double epsilon = 0.0001;

	@Override
	public void init(Map<String, String> params, ParserConfiguration config) {
		String epsilonString = params.get(EPSILON_NAME);
		if (epsilonString != null) {
			epsilon = Double.parseDouble(epsilonString);
		}
	}

	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}
	
	@Override
	public Set<Requirement> appliesToRequirements() {
		return CollectionUtils.asSet(Requirement.R_GE_R_CONSECUTIVE_POINTS_SAME, Requirement.R_GE_R_SELF_INTERSECTION);
	}

	@Override
	public void check(LinearRing lr) {
		List<Vertex> pointList = lr.getVertices();

		// check for consecutive points first
		for (int i = 0; i < pointList.size() - 1; i++) {
			Vertex point1 = pointList.get(i);
			Vertex point2 = pointList.get(i + 1);
			if (point1.equalsWithEpsilon(point2, epsilon)) {
				// consecutive points same
				CheckError err = new ConsecutivePointSameError(lr, point1, point2);
				CheckResult cr = new CheckResult(this, ResultStatus.ERROR, err);
				lr.addCheckResult(cr);
				return;
			}
		}

		// ignore last point, because last point = first, but this is allowed
		for (int i = 0; i < pointList.size() - 2; i++) {
			Vertex point1 = pointList.get(i);
			for (int j = i + 2; j < pointList.size() - 1; j++) {
				Vertex point2 = pointList.get(j);
				if (point1.equalsWithEpsilon(point2, epsilon)) {
					// non consecutive points same
					CheckError err = new RingDuplicatePointError(lr, point1, point2);
					CheckResult cr = new CheckResult(this, ResultStatus.ERROR, err);
					lr.addCheckResult(cr);
					return;
				}
			}
		}
		CheckResult cr = new CheckResult(this, ResultStatus.OK, null);
		lr.addCheckResult(cr);
	}

	@Override
	public RequirementType getType() {
		return RequirementType.GEOMETRY;
	}

	@Override
	public Check createNewInstance() {
		return new DuplicatePointsCheck();
	}

	@Override
	public CheckId getCheckId() {
		return CheckId.C_GE_R_DUPLICATE_POINT;
	}

}
