/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.RequirementType;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.error.NestedRingError;
import de.hft.stuttgart.citydoctor2.checks.util.CollectionUtils;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

/**
 * Checks whether a inner ring is completely contained in another inner ring
 * within the same polygon
 * 
 * @author Matthias Betz
 *
 */
public class NestedRingsCheck extends Check {

	private static final List<CheckId> dependencies;

	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		deps.add(CheckId.C_GE_R_TOO_FEW_POINTS);
		deps.add(CheckId.C_GE_R_NOT_CLOSED);
		deps.add(CheckId.C_GE_R_DUPLICATE_POINT);
		deps.add(CheckId.C_GE_R_SELF_INTERSECTION);
		deps.add(CheckId.C_GE_P_NON_PLANAR);
		deps.add(CheckId.C_GE_P_ORIENTATION_RINGS_SAME);
		dependencies = Collections.unmodifiableList(deps);
	}

	@Override
	public void check(Polygon p) {
		for (LinearRing interiorRing : p.getInnerRings()) {
			for (LinearRing checkRing : p.getInnerRings()) {
				if (checkRing == interiorRing) {
					// do not compare with itself
					continue;
				}
				// are all points from checkRing inside interiorRing?
				if (areAllPointsInside(interiorRing, checkRing)) {
					// found all vertices inside another interior ring
					CheckError err = new NestedRingError(p, interiorRing, checkRing);
					CheckResult cr = new CheckResult(this, ResultStatus.ERROR, err);
					p.addCheckResult(cr);
					return;
				}
			}
		}
		CheckResult cr = new CheckResult(this, ResultStatus.OK, null);
		p.addCheckResult(cr);
	}

	private boolean areAllPointsInside(LinearRing ring, LinearRing checkRing) {
		boolean isInside = true;
		for (Vertex v : checkRing.getVertices()) {
			if (!ring.isPointInside(v)) {
				isInside = false;
				break;
			}
		}
		return isInside;
	}

	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}
	
	@Override
	public Set<Requirement> appliesToRequirements() {
		return CollectionUtils.singletonSet(Requirement.R_GE_P_INNER_RINGS_NESTED);
	}

	@Override
	public RequirementType getType() {
		return RequirementType.GEOMETRY;
	}

	@Override
	public Check createNewInstance() {
		return new NestedRingsCheck();
	}

	@Override
	public CheckId getCheckId() {
		return CheckId.C_GE_P_INNER_RINGS_NESTED;
	}

}
