/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.semantics;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.RequirementType;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.error.NotFloorError;
import de.hft.stuttgart.citydoctor2.checks.util.CollectionUtils;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurfaceType;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.math.Vector3d;


/**
 * SEM_ISCEILING checks if the normal direction of a polygon labeled as
 * OuterCeilingSurface points downwards.
 * 
 * @author Matthias Betz - 12bema1bif@hft-stuttgart.de
 *
 */
public class IsFloorCheck extends Check {
	
	private static final List<CheckId> dependencies;
	
	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		deps.add(CheckId.C_GE_R_TOO_FEW_POINTS);
		deps.add(CheckId.C_GE_R_NOT_CLOSED);
		deps.add(CheckId.C_GE_R_DUPLICATE_POINT);
		deps.add(CheckId.C_GE_R_SELF_INTERSECTION);
		deps.add(CheckId.C_GE_P_NON_PLANAR);
		dependencies = Collections.unmodifiableList(deps);
	}
	
	@Override
	public void check(BoundarySurface bs) {
		// works only on outer floor polygons
		if (bs.getType() != BoundarySurfaceType.OUTER_FLOOR && bs.getType() != BoundarySurfaceType.FLOOR) {
			return;
		}
		for (Geometry geom : bs.getGeometries()) {
			for (Polygon p : geom.getPolygons()) {
				Vector3d normal = p.calculateNormalNormalized();
				if (normal.getZ() <= 0) {
					NotFloorError err = new NotFloorError(bs, p);
					CheckResult cr = new CheckResult(this, ResultStatus.ERROR, err);
					bs.addCheckResult(cr);
					return;
				}
			}
		}
		CheckResult cr = new CheckResult(this, ResultStatus.OK, null);
		bs.addCheckResult(cr);
	}
	
	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}
	
	@Override
	public Set<Requirement> appliesToRequirements() {
		return CollectionUtils.singletonSet(Requirement.R_SE_BS_IS_FLOOR);
	}
	
	@Override
	public RequirementType getType() {
		return RequirementType.SEMANTIC;
	}

	@Override
	public Check createNewInstance() {
		return new IsFloorCheck();
	}

	@Override
	public CheckId getCheckId() {
		return CheckId.C_SE_BS_IS_FLOOR;
	}

}
