/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import static org.junit.Assert.assertEquals;

import org.junit.Test;

import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

/**
 * 
 * @author Matthias Betz
 *
 */
public class PolygonSelfIntCheckTest {

	
	
	@Test
	public void testPolygonInteriorIntersectExterior() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		
		Polygon poly = new ConcretePolygon();
		geom.addPolygon(poly);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		lr.setParent(poly);
		poly.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		Vertex v1 = new Vertex(10, 0, 0);
		lr.addVertex(v1);

		Vertex v2 = new Vertex(10, 10, 0);
		lr.addVertex(v2);

		Vertex v3 = new Vertex(0, 10, 0);
		lr.addVertex(v3);
		
		lr.addVertex(v0);
		
		LinearRing interior = new LinearRing(LinearRingType.INTERIOR);
		poly.addInteriorRing(interior);
		
		Vertex v4 = new Vertex(8, 2, 0);
		interior.addVertex(v4);
		
		Vertex v5 = new Vertex(8, 4, 0);
		interior.addVertex(v5);

		Vertex v6 = new Vertex(12, 4, 0);
		interior.addVertex(v6);

		Vertex v7 = new Vertex(12, 2, 0);
		interior.addVertex(v7);
		interior.addVertex(v4);

		PolygonIntersectingRingsCheck c = new PolygonIntersectingRingsCheck();
		
		c.check(poly);
		
		CheckResult cr = poly.getCheckResult(c);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_P_INTERSECTING_RINGS, cr.getError().getErrorId());
	}
	
	@Test
	public void testPolygonGood() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		
		Polygon poly = new ConcretePolygon();
		geom.addPolygon(poly);
			
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		poly.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		Vertex v1 = new Vertex(10, 0, 0);
		lr.addVertex(v1);

		Vertex v2 = new Vertex(10, 10, 0);
		lr.addVertex(v2);

		Vertex v3 = new Vertex(0, 10, 0);
		lr.addVertex(v3);
		
		lr.addVertex(v0);
		
		LinearRing interior = new LinearRing(LinearRingType.INTERIOR);
		interior.setParent(poly);
		poly.addInteriorRing(interior);
		
		Vertex v4 = new Vertex(8, 2, 0);
		interior.addVertex(v4);
		
		Vertex v5 = new Vertex(8, 4, 0);
		interior.addVertex(v5);

		Vertex v6 = new Vertex(9, 4, 0);
		interior.addVertex(v6);

		Vertex v7 = new Vertex(9, 2, 0);
		interior.addVertex(v7);
		interior.addVertex(v4);

		PolygonIntersectingRingsCheck c = new PolygonIntersectingRingsCheck();
		
		c.check(poly);
		
		assertEquals(ResultStatus.OK, poly.getCheckResult(c).getResultStatus());
	}
	
	@Test
	public void test2InteriorRingsIntersecting() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		
		Polygon poly = new ConcretePolygon();
		geom.addPolygon(poly);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		lr.setParent(poly);
		poly.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		Vertex v1 = new Vertex(10, 0, 0);
		lr.addVertex(v1);

		Vertex v2 = new Vertex(10, 10, 0);
		lr.addVertex(v2);

		Vertex v3 = new Vertex(0, 10, 0);
		lr.addVertex(v3);
		
		lr.addVertex(v0);
		
		LinearRing interior = new LinearRing(LinearRingType.INTERIOR);
		interior.setParent(poly);
		poly.addInteriorRing(interior);
		
		Vertex v4 = new Vertex(5, 2, 0);
		interior.addVertex(v4);
		
		Vertex v5 = new Vertex(5, 4, 0);
		interior.addVertex(v5);

		Vertex v6 = new Vertex(9, 4, 0);
		interior.addVertex(v6);

		Vertex v7 = new Vertex(9, 2, 0);
		interior.addVertex(v7);
		interior.addVertex(v4);
		
		LinearRing interior2 = new LinearRing(LinearRingType.INTERIOR);
		interior2.setParent(poly);
		poly.addInteriorRing(interior2);
		
		Vertex v8 = new Vertex(3, 2, 0);
		interior2.addVertex(v8);
		
		Vertex v9 = new Vertex(3, 3, 0);
		interior2.addVertex(v9);

		Vertex v10 = new Vertex(6, 3, 0);
		interior2.addVertex(v10);

		Vertex v11 = new Vertex(6, 2, 0);
		interior2.addVertex(v11);
		interior2.addVertex(v8);

		PolygonIntersectingRingsCheck c = new PolygonIntersectingRingsCheck();
		
		c.check(poly);
		
		CheckResult cr = poly.getCheckResult(c);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_P_INTERSECTING_RINGS, cr.getError().getErrorId());
	}
	
	@Test
	public void test2InteriorRingsNotIntersecting() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		
		Polygon poly = new ConcretePolygon();
		geom.addPolygon(poly);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		poly.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		Vertex v1 = new Vertex(10, 0, 0);
		lr.addVertex(v1);

		Vertex v2 = new Vertex(10, 10, 0);
		lr.addVertex(v2);

		Vertex v3 = new Vertex(0, 10, 0);
		lr.addVertex(v3);
		
		lr.addVertex(v0);
		
		LinearRing interior = new LinearRing(LinearRingType.INTERIOR);
		poly.addInteriorRing(interior);
		
		Vertex v4 = new Vertex(5, 2, 0);
		interior.addVertex(v4);
		
		Vertex v5 = new Vertex(5, 4, 0);
		interior.addVertex(v5);

		Vertex v6 = new Vertex(9, 4, 0);
		interior.addVertex(v6);

		Vertex v7 = new Vertex(9, 2, 0);
		interior.addVertex(v7);
		interior.addVertex(v4);
		
		LinearRing interior2 = new LinearRing(LinearRingType.INTERIOR);
		poly.addInteriorRing(interior2);
		
		Vertex v8 = new Vertex(3, 2, 0);
		interior2.addVertex(v8);
		
		Vertex v9 = new Vertex(3, 3, 0);
		interior2.addVertex(v9);

		Vertex v10 = new Vertex(4, 3, 0);
		interior2.addVertex(v10);

		Vertex v11 = new Vertex(4, 2, 0);
		interior2.addVertex(v11);
		interior2.addVertex(v8);

		PolygonIntersectingRingsCheck c = new PolygonIntersectingRingsCheck();
		
		c.check(poly);
		
		assertEquals(ResultStatus.OK, poly.getCheckResult(c).getResultStatus());
	}

}
